@@ -1,10 +1,15 @@
-package Pcore::API::Whois v0.1.3;
+package Pcore::API::Whois v0.2.0;
 
 use Pcore -dist, -class;
 use Pcore::API::Whois::Server;
 use Pcore::API::Whois::Response qw[:CONST];
 use Net::Whois::Raw::Data qw[];
 
+has cache => ( is => 'ro', isa => Str );                                             # path to whois cache db
+has cache_timeout => ( is => 'ro', isa => PositiveInt, default => 60 * 60 * 24 );    # in seconds
+
+has dbh => ( is => 'lazy', isa => Maybe [ InstanceOf ['Pcore::Handle::SQLite'] ], init_arg => undef );
+
 our $DB = P->cfg->load( $ENV->share->get('/data/whois.perl') );
 
 our $PUB_SUFFIX;
@@ -17,9 +22,47 @@ for my $pub_suffix ( keys $DB->{pub_suffix_server}->%* ) {
     $PUB_SUFFIX->{ P->host($pub_suffix)->name } = $DB->{pub_suffix_server}->{$pub_suffix};
 }
 
+sub _build_dbh ($self) {
+    return if !$self->cache;
+
+    H->add(
+        whois        => 'SQLite',
+        addr         => 'file:' . $self->cache,
+        journal_mode => 'WAL',
+        synchronous  => 'OFF',
+        cache_size   => 20_000,
+    );
+
+    # DDL
+    my $ddl = H->whois->ddl;
+
+    $ddl->add_changeset(
+        id  => 1,
+        sql => <<'SQL'
+                CREATE TABLE IF NOT EXISTS `whois` (
+                    `domain` TEXT PRIMARY KEY NOT NULL,
+                    `last_checked` INTEGER NULL,
+                    `status` INTEGER NOT NULL
+                );
+
+                CREATE INDEX IF NOT EXISTS `idx_whois_domain_last_checked` ON `whois` (`domain` ASC, `last_checked` ASC);
+SQL
+    );
+
+    $ddl->upgrade;
+
+    return H->whois;
+}
+
 sub search ( $self, $domain, $cb = undef ) {
     my $blocking_cv = defined wantarray ? AE::cv : undef;
 
+    state $q1 = $self->cache ? $self->dbh->query('SELECT status FROM whois WHERE domain = ? AND last_checked > ?') : undef;
+
+    state $q2 = $self->cache ? $self->dbh->query('INSERT OR IGNORE INTO whois (last_checked, status, domain) VALUES (?, ?, ?)') : undef;
+
+    state $q3 = $self->cache ? $self->dbh->query('UPDATE whois SET last_checked = ?, status = ? WHERE domain = ?') : undef;
+
     my $server_host;
 
     my @label = split /[.]/sm, P->host($domain)->canon;
@@ -44,7 +87,8 @@ sub search ( $self, $domain, $cb = undef ) {
 
     if ( !$server_host ) {
         my $res = Pcore::API::Whois::Response->new(
-            {   status => $WHOIS_NOT_SUPPORTED,
+            {   query  => $query,
+                status => $WHOIS_NOT_SUPPORTED,
                 reason => $Pcore::API::Whois::Response::STATUS_REASON->{$WHOIS_NOT_SUPPORTED}
             }
         );
@@ -54,12 +98,33 @@ sub search ( $self, $domain, $cb = undef ) {
         $blocking_cv->($res) if $blocking_cv;
     }
     else {
+        if ( $self->cache ) {
+
+            # check cache and return cached response
+            if ( my $status = $q1->selectval( bind => [ $query, time - $self->cache_timeout ] ) ) {
+                my $res = Pcore::API::Whois::Response->new(
+                    {   query     => $query,
+                        is_cached => 1,
+                        status    => $status->$*,
+                        reason    => $Pcore::API::Whois::Response::STATUS_REASON->{ $status->$* }
+                    }
+                );
+
+                $cb->($res) if $cb;
+
+                $blocking_cv->($res) if $blocking_cv;
+
+                return;
+            }
+        }
+
         Pcore::API::Whois::Server->new(
             {   host => $server_host,
                 cb   => sub($server) {
                     if ( !$server || !$server->is_supported ) {
                         my $res = Pcore::API::Whois::Response->new(
-                            {   status => $WHOIS_NOT_SUPPORTED,
+                            {   query  => $query,
+                                status => $WHOIS_NOT_SUPPORTED,
                                 reason => $Pcore::API::Whois::Response::STATUS_REASON->{$WHOIS_NOT_SUPPORTED}
                             }
                         );
@@ -98,6 +163,9 @@ sub search ( $self, $domain, $cb = undef ) {
                                             }
                                         }
                                     }
+
+                                    # cache results for successful query
+                                    $q2->do( bind => [ time, $res->status, $query ] ) || $q3->do( bind => [ time, $res->status, $query ] ) if $self->cache;
                                 }
 
                                 $cb->($res) if $cb;
@@ -158,11 +226,11 @@ sub export_supported_pub_suffixes ($self) {
 ## +------+----------------------+----------------------------------------------------------------------------------------------------------------+
 ## | Sev. | Lines                | Policy                                                                                                         |
 ## |======+======================+================================================================================================================|
-## |    3 | 16, 126, 143, 147    | References::ProhibitDoubleSigils - Double-sigil dereference                                                    |
+## |    3 | 21, 194, 211, 215    | References::ProhibitDoubleSigils - Double-sigil dereference                                                    |
 ## |------+----------------------+----------------------------------------------------------------------------------------------------------------|
-## |    3 | 20                   | Subroutines::ProhibitExcessComplexity - Subroutine "search" with high complexity score (27)                    |
+## |    3 | 57                   | Subroutines::ProhibitExcessComplexity - Subroutine "search" with high complexity score (36)                    |
 ## |------+----------------------+----------------------------------------------------------------------------------------------------------------|
-## |    3 | 80, 92               | ControlStructures::ProhibitDeepNests - Code structure is deeply nested                                         |
+## |    3 | 145, 157             | ControlStructures::ProhibitDeepNests - Code structure is deeply nested                                         |
 ## +------+----------------------+----------------------------------------------------------------------------------------------------------------+
 ##
 ## -----SOURCE FILTER LOG END-----
